import pickle
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler
import pandas as pd

# ----------------- #
# load model 
# ----------------- #
def load_model(path: str):
    """
    Parameters:
    :path - Path to the model saved in .pkl
    ---------------------------
    The function load the model saved previously in the pickle format.
    """
    with open(path, 'rb') as file:
        model = pickle.load(file)
    return model 


# ----------------- #
# setup data for train 
# ----------------- #
def setup(X, y, dummy: bool = False, seed: int = 42, size: float = 0.25):
    """
    Parameters:
    :X - The model matrix X (features matrix)
    :y - The target variable 
    :dummy - If False, the function produces the OHE. If True, the dummy encoding 
    :seed - Random seed to apply at the train_test_split function 
    :size - Size of the test dataset 
    ---------------------------
    The function returns the following elements: X_train, X_test, y_train, y_test.
    """
    # split dataset in train and test
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=size, random_state=seed)
    X_train = pd.get_dummies(data=X_train, drop_first=dummy)
    X_test = pd.get_dummies(data=X_test, drop_first=dummy)
    return X_train, y_train, X_test, y_test


# --------------------- #
# scale X dataset
# --------------------- #
def scaling(X):
    """
    Parameters:
    :X - The model matrix X/X_train/X_test
    ---------------------------
    The function returns the normalised matrix.
    """
    # dataframe check 
    dataframe_review(X)
    # scaling quantitative variables 
    types = X.dtypes
    quant = types[types != 'object']
    quant_columns = list(quant.index)
    scaler = StandardScaler()
    scaler.fit(X[quant_columns])
    X[quant_columns] = scaler.transform(X[quant_columns])
    return X


# ----------------- #
# OHE check
# ----------------- #
def dummy_control(data):
    """
    Parameters:
    :data - A Pandas dataframe 
    ---------------------------
    The function checks if the Pandas dataframe passed is encoded with dummy or OHE. 
    """
    types = data.dtypes
    qual_col = types[types == 'object']
    if len(qual_col) != 0:
        raise TypeError('dataframe with non-numerical columns')
    else:
        pass


# ----------------- #
#  pandas dataframe check
# ----------------- #
def dataframe_review(data) -> None:
    """
    Parameters:
    :data - A dataframe 
    ---------------------------
    The function checks if the object passed is a Pandas dataframe.
    """
    if data.__class__.__name__ == 'DataFrame':
        pass
    else:
        raise TypeError('The data loaded is not a DataFrame')
    


import pandas as pd 
from IPython.display import display
import numpy as np 
import xgboost as xgb
import matplotlib.pyplot as plt 



# ------------------- #
#     LINEAR MODEL 
# ------------------- #
def check_linear(model):
    """
    Parameters:
    :model - Sklearn model 
    ---------------------------
    The function checks if the model passed is a linear regression.
    """
    if model.__class__.__name__ != 'LinearRegression':
        raise TypeError('The model passed isn\'t a linear model')
        

def linear_coef(model):
    """
    Parameters:
    :model - A linear regression model
    ---------------------------
    The function produces a dataframe that contains the values of every feature of the regression matrix X passed to the linear model.
    """
    check_linear(model)
    intercept = ('intercept',model.intercept_)
    coef = list(zip(list(model.feature_names_in_), model.coef_))
    coef = [intercept, *coef]
    df_coef = pd.DataFrame(coef)
    df_coef.columns = ['Columns', 'Linear Coef']
    # display step 
    display(df_coef)


# ------------------- #
#     LASSO MODEL 
# ------------------- #
def check_lasso(model): 
    """
    Parameters:
    :model - Sklearn model 
    ---------------------------
    The function checks if the model passed is a lasso regression.
    """
    if model.__class__.__name__ != 'Lasso':
        raise TypeError('The model passed isn\'t a lasso model')

def lasso_coef(model):
    """
    Parameters:
    :model - A lasso regression model
    ---------------------------
    The function produces a dataframe that contains the values of every feature of the regression matrix X passed to the lasso model.
    """
    check_lasso(model)
    intercept = ('intercept',model.intercept_)
    coef = list(zip(list(model.feature_names_in_), model.coef_))
    coef = [intercept, *coef]
    df_coef = pd.DataFrame(coef)
    df_coef.columns = ['Columns', 'Lasso Coef']
    # display step 
    display(df_coef)



# ------------------- #
#     RIDGE MODEL 
# ------------------- #
def check_ridge(model): 
    """
    Parameters:
    :model - Sklearn model 
    ---------------------------
    The function checks if the model passed is a ridge regression.
    """
    if model.__class__.__name__ != 'Ridge':
        raise TypeError('The model passed isn\'t a ridge model')


def ridge_coef(model):
    """
    Parameters:
    :model - A lasso regression model
    ---------------------------
    The function produces a dataframe that contains the values of every feature of the regression matrix X passed to the ridge model.
    """
    check_ridge(model)
    intercept = ('intercept',model.intercept_)
    coef = list(zip(list(model.feature_names_in_), model.coef_))
    coef = [intercept, *coef]
    df_coef = pd.DataFrame(coef)
    df_coef.columns = ['Columns', 'Ridge Coef']
    # display step 
    display(df_coef)


# ------------------- #
#     RANDOM FOREST 
# ------------------- #
def check_random_forest(model): 
    """
    Parameters:
    :model - Sklearn model 
    ---------------------------
    The function checks if the model passed is a random forest regression.
    """
    if model.__class__.__name__ != 'RandomForestRegressor':
        raise TypeError('The model passed isn\'t a ridge model')


def random_forest_fi(model, figsize: tuple[float, float] = (12,10)):
    """
    Parameters:
    :model - A RF regression model
    ---------------------------
    The function displays the feature importance plot. 
    """
    check_random_forest(model)
    importances = model.feature_importances_
    std = np.std([tree.feature_importances_ for tree in model.estimators_], axis=0)
    feature_names = model.feature_names_in_
    forest_importances = pd.Series(importances, index=feature_names)
    plt.figure(figsize=figsize)
    forest_importances.plot.bar(yerr=std)
    plt.title("Feature importances using mean decrease in impurity")
    plt.ylabel("Mean decrease in impurity")
    plt.show()


# ------------------- #
#     XGBOOST 
# ------------------- #
def check_xgboost(model): 
    """
    Parameters:
    :model - xgboost model
    ---------------------------
    The function checks if the model passed is a xgboost regression.
    """
    if model.__class__.__name__ != 'XGBRegressor':
        raise TypeError('The model passed isn\'t an xgboost')


def xgboost_fi(model, figsize: tuple[float, float] = (12,10)):
    """
    Parameters:
    :model - A xgboost regression model
    ---------------------------
    The function displays the feature importance plot. 
    """
    check_xgboost(model)
    xgb.plot_importance(model)
    plt.rcParams['figure.figsize'] = [figsize[0], figsize[1]]
    plt.show()


import pandas as pd 
from IPython.display import display
import numpy as np 
import matplotlib.pyplot as plt 
from yellowbrick.regressor import prediction_error


class Diagnose():

    def __init__(self, model, X_train, y_train, X_test, y_test):
        self.model = model 
        self.X_train = X_train
        self.y_train = y_train
        self.X_test = X_test
        self.y_test = y_test


    def check_model(self, model_name: str) -> None:
        if self.model.__class__.__name__ != model_name:
            raise TypeError('The model passed isn\'t a linear model')


    def coef(self): 
        intercept = ('intercept', self.model.intercept_)
        coef = list(zip(list(self.model.feature_names_in_), self.model.coef_))
        coef = [intercept, *coef]
        df_coef = pd.DataFrame(coef)
        df_coef.columns = ['Columns', 'Coef']
        # display step 
        display(df_coef)
    

    def model_predict_error(self):
        visualizer = prediction_error(self.model, self.X_train, self.y_train, self.X_test, self.y_test)


class LinearDiagnose(Diagnose):

    def __init__(self, model, X_train, y_train, X_test, y_test):
        super().__init__(model, X_train, y_train, X_test, y_test)


    def check_model(self, model_name: str = "LinearRegression") -> None:
        return super().check_model(model_name)


class LassoDiagnose(Diagnose):

    def __init__(self, model, X_train, y_train, X_test, y_test):
        super().__init__(model, X_train, y_train, X_test, y_test)


    def check_model(self, model_name: str = "Lasso") -> None:
        return super().check_model(model_name)


class RidgeDiagnose(Diagnose):

    def __init__(self, model, X_train, y_train, X_test, y_test):
        super().__init__(model, X_train, y_train, X_test, y_test)


    def check_model(self, model_name: str = "Ridge") -> None:
        return super().check_model(model_name)
    


